// ===================================================================
// ====   BALM10beep    générateur UHF , balise  pour manips      ====
// ====   ballon sonde météo M10, pll ADF7012B,GPS Sierra, Nano   ==== 
// ====   version 1.0                   F5BEZ Dominique LEVEQUE   ====
// ===================================================================
// prérequit : 
//  liens https://github.com   voir à la fin de ce fichier  copyright (C) 2012  KT5TK
//  datasheet ADF701 . pdf     dont détails et rôles des régistres 0 à 3 

#include "BALM10beep.h"
#include "BALbgestionbeep.h"
#include "Balbadf7012.h"

#define SPI_BITBANG   // déjà défini avec BALM10beep.h
#include <math.h>

#if !defined(SPI_BITBANG)
#include <SPI.h>
#endif

#if defined(ARDUINO) && ARDUINO >= 100
#include <Arduino.h>
#else
#include <WProgram.h>
#endif

int Mycorrectqrg= 90;   //       personnalisation selon   M10 et son xtal 8Mhz

const int MAX_RES = 16;
char res_adf7012[MAX_RES];
unsigned int powerlevel;

// Configuration storage structs =============================================
struct {
    struct {
        unsigned int  frequency_error_correction;
        unsigned char r_divider;
        unsigned char crystal_doubler;
        unsigned char crystal_oscillator_disable;
        unsigned char clock_out_divider;
        unsigned char vco_adjust;
        unsigned char output_divider;
    } r0;

    struct {
        unsigned int  fractional_n;
        unsigned char integer_n;
        unsigned char prescaler;
    } r1;

    struct {
        unsigned char mod_control;
        unsigned char gook;
        unsigned char power_amplifier_level;
        unsigned int  modulation_deviation;
        unsigned char gfsk_modulation_control;
        unsigned char index_counter;
    } r2;

    struct {
        unsigned char pll_enable;
        unsigned char pa_enable;
        unsigned char clkout_enable;
        unsigned char data_invert;
        unsigned char charge_pump_current;
        unsigned char bleed_up;
        unsigned char bleed_down;
        unsigned char vco_disable;
        unsigned char muxout;
        unsigned char ld_precision;
        unsigned char vco_bias;
        unsigned char pa_bias;
        unsigned char pll_test_mode;
        unsigned char sd_test_mode;
    } r3;
} adf_config;


// Write directly to AVR port in SPIwrite() instead of using digitalWrite()
//#define FAST_IO

// Configuration functions ===================================================

// Config resetting functions --------------------------------------------
void RadioAdf7012::adf_reset_config(void) 
{
  
    adf_reset_register_zero();
    adf_reset_register_one();
    adf_reset_register_two();
    adf_reset_register_three();
    adf_reset();  

//    while(!adf_reg_ready());

}
//--------------------------------------------------------------------
// Power up default settings are defined here:

void RadioAdf7012::adf_reset_register_zero(void) 
{
    adf_config.r0.frequency_error_correction = 0;               // Don't bother for now...
    adf_config.r0.r_divider = ADF7012_CRYSTAL_DIVIDER;          // Whatever works best for 2m, 1.25m and 70 cm ham bands
    adf_config.r0.crystal_doubler = 0;                          // Who would want that? Lower f_pfd means finer channel steps.
    adf_config.r0.crystal_oscillator_disable = 1;               // Disable internal crystal oscillator because we have an external VCXO
    adf_config.r0.clock_out_divider = 2;                        // Don't bother for now...
    adf_config.r0.vco_adjust = 0;                               // Don't bother for now... (Will be automatically adjusted until PLL lock is achieved)
    adf_config.r0.output_divider = ADF_OUTPUT_DIVIDER_BY_4;     // Pre-set div 4 for 2m. Will be changed according tx frequency on the fly
}
//--------------------------------------------------------------------
void RadioAdf7012::adf_reset_register_one(void) 
{
    adf_config.r1.integer_n = 50;                              // Pre-set for 400M avec xtal 8M
    adf_config.r1.fractional_n = 0;                            // Pre-set for 400M
    adf_config.r1.prescaler = ADF_PRESCALER_8_9;                // 8/9 requires an integer_n > 91; 4/5 only requires integer_n > 31
}
//--------------------------------------------------------------------
void RadioAdf7012::adf_reset_register_two(void) 
{
   adf_config.r2.mod_control = ADF_MODULATION_ASK;               // For AFSK the modulation is done through the external VCXO we don't want any FM generated by the ADF7012 itself
 //  adf_config.r2.mod_control = ADF_MODULATION_FSK;             // bez test  pb  déjà mod fm 1M sans signal data
 //  adf_config.r2.mod_control = ADF_MODULATION_GFSK;            // bez       nb si gfsk et gook il faut inp clockdatamodul 
 //  adf_config.r2.mod_control = ADF_MODULATION_OOK;             // bez fait tou ou rien selon data en level
    adf_config.r2.gook = 0;                                     // Whatever... This might give us a nicer swing in phase maybe...
    adf_config.r2.power_amplifier_level = 32;                   // 32  16??? is about half maximum power. Output −20dBm at 0x0, and 13 dBm at 0x7E at 868 MHz   // #define MAXINDICEPWR 63 // 2p6 -1  reg 2 P1..P6
  // def AM
 //   adf_config.r2.modulation_deviation = 16;                    // 16 is about half maximum amplitude @ ASK.
  // FM
       adf_config.r2.modulation_deviation = 6;                    // 16 is about half maximum amplitude @ ASK.
    adf_config.r2.gfsk_modulation_control = 0;                  // Don't bother for now...
    adf_config.r2.index_counter = 0;                            // Don't bother for now...
}
//--------------------------------------------------------------------
void RadioAdf7012::adf_reset_register_three(void) 
{
    adf_config.r3.pll_enable = 0;                               // Switch off PLL (will be switched on after Ureg is checked and confirmed ok)
    adf_config.r3.pa_enable = 0;                                // Switch off PA  (will be switched on when PLL lock is confirmed)
    adf_config.r3.clkout_enable = 0;                            // No clock output needed at the moment
    adf_config.r3.data_invert = 1;                              // Results in a TX signal when TXDATA input is low
    adf_config.r3.charge_pump_current = ADF_CP_CURRENT_2_1;     // 2.1 mA. This is the maximum
    adf_config.r3.bleed_up = 0;                                 // Don't worry, be happy...
    adf_config.r3.bleed_down = 0;                               // Dito
    adf_config.r3.vco_disable = 0;                              // VCO is on
    
    adf_config.r3.muxout = ADF_MUXOUT_REG_READY;                // Lights up the green LED if the ADF7012 is properly powered (changes to lock detection in a later stage)

    adf_config.r3.ld_precision = ADF_LD_PRECISION_3_CYCLES;     // What the heck? It is recommended that LDP be set to 1; 0 is more relaxed
    adf_config.r3.vco_bias = 6;                                 // In 0.5 mA steps; Default 6 means 3 mA; Maximum (15) is 8 mA
    adf_config.r3.pa_bias = 4;                                  // In 1 mA steps; Default 4 means 8 mA; Minimum (0) is 5 mA; Maximum (7) is 12 mA (Datasheet says uA which is bullshit)
    adf_config.r3.pll_test_mode = 0;
    adf_config.r3.sd_test_mode = 0;
}

//-----------------------------------------------------------
void RadioAdf7012::adf_reset(void) 
{
    digitalWrite(SSpin,   HIGH);
    digitalWrite(ADF7012_TX_DATA_PIN, HIGH);
    digitalWrite(SCKpin,  HIGH);
    digitalWrite(MOSIpin, HIGH);
    delay(100);
}

//-----------------------------------------------------------
// Configuration writing functions --------------------------
void RadioAdf7012::adf_write_config(void) 
{
    adf_write_register_zero();
    adf_write_register_one();
    adf_write_register_two();
    adf_write_register_three();
}
//---------------------------------------------------------------------------------------------------------------
// les valeurs par défaut ou modifiées sont masquées ( & ET bit )  pour ne garder que les bits utiles
// puis décalèes à gauche ( << ) de x places puis assemblées ( | ou bit )  pour former le mot vers le registre (r)  
//----------------------------------------------------------------------------------------------------------------
void RadioAdf7012::adf_write_register_zero(void) 
{
    unsigned long reg =
        (0) |
        ((unsigned long)(adf_config.r0.frequency_error_correction & 0x7FF) << 2U) |
        ((unsigned long)(adf_config.r0.r_divider & 0xF ) << 13U) |
        ((unsigned long)(adf_config.r0.crystal_doubler & 0x1 ) << 17U) |
        ((unsigned long)(adf_config.r0.crystal_oscillator_disable & 0x1 ) << 18U) |
        ((unsigned long)(adf_config.r0.clock_out_divider & 0xF ) << 19U) |
        ((unsigned long)(adf_config.r0.vco_adjust & 0x3 ) << 23U) |
        ((unsigned long)(adf_config.r0.output_divider & 0x3 ) << 25U);
     
    adf_write_register(reg);
}

//-----------------------------------------------------------
void RadioAdf7012::adf_write_register_one(void) 
{
    unsigned long reg =
        (1) |
        ((unsigned long)(adf_config.r1.fractional_n & 0xFFF) << 2) |
        ((unsigned long)(adf_config.r1.integer_n & 0xFF ) << 14) |
        ((unsigned long)(adf_config.r1.prescaler & 0x1 ) << 22);

    adf_write_register(reg);
}

//-----------------------------------------------------------
void RadioAdf7012::adf_write_register_two(void) 
{
    unsigned long reg =
        (2) |
        ((unsigned long)(adf_config.r2.mod_control & 0x3 ) << 2) |
        ((unsigned long)(adf_config.r2.gook & 0x1 ) << 4) |
        ((unsigned long)(adf_config.r2.power_amplifier_level & 0x3F ) << 5) |
        ((unsigned long)(adf_config.r2.modulation_deviation & 0x1FF) << 11) |
        ((unsigned long)(adf_config.r2.gfsk_modulation_control & 0x7 ) << 20) |
        ((unsigned long)(adf_config.r2.index_counter & 0x3 ) << 23);

    adf_write_register(reg);
}

//-----------------------------------------------------------
void RadioAdf7012::adf_write_register_three(void) 
{
    unsigned long reg =
        (3) |
        ((unsigned long)(adf_config.r3.pll_enable & 0x1 ) << 2) |
        ((unsigned long)(adf_config.r3.pa_enable & 0x1 ) << 3) |
        ((unsigned long)(adf_config.r3.clkout_enable & 0x1 ) << 4) |
        ((unsigned long)(adf_config.r3.data_invert & 0x1 ) << 5) |
        ((unsigned long)(adf_config.r3.charge_pump_current & 0x3 ) << 6) |
        ((unsigned long)(adf_config.r3.bleed_up & 0x1 ) << 8) |
        ((unsigned long)(adf_config.r3.bleed_down & 0x1 ) << 9) |
        ((unsigned long)(adf_config.r3.vco_disable & 0x1 ) << 10) |
        ((unsigned long)(adf_config.r3.muxout & 0xF ) << 11) |
        ((unsigned long)(adf_config.r3.ld_precision & 0x1 ) << 15) |
        ((unsigned long)(adf_config.r3.vco_bias & 0xF ) << 16) |
        ((unsigned long)(adf_config.r3.pa_bias & 0x7 ) << 20) |
        ((unsigned long)(adf_config.r3.pll_test_mode & 0x1F ) << 23) |
        ((unsigned long)(adf_config.r3.sd_test_mode & 0xF ) << 28);
        
    adf_write_register(reg);
}

//-----------------------------------------------------------
void RadioAdf7012::adf_write_register(unsigned long data) 
{
  //adf_reset();
  digitalWrite(SSpin,   HIGH);
  //!!!!digitalWrite(ADF7012_TX_DATA_PIN, HIGH);  //org  mais pas utile et sinon modifie ensuite les réglages 
  digitalWrite(SCKpin,  HIGH);
  digitalWrite(MOSIpin, HIGH);
  
#if !defined(SPI_BITBANG)   // use SPI library
  // take the SS pin low to select the ADF7012 chip:
  digitalWrite(SSpin,LOW);
  
  for (int j = 3; j >= 0; j--) // Loop through the 4 bytes of the unsigned long
  {
    byte wordb = (byte) (data >> (j * 8));
    SPI.transfer(wordb);
    Serial.print(wordb, HEX);
    Serial.print("j");
     delay(1000);
  } 

  // take the SS pin high to de-select the ADF7012 chip:
  digitalWrite(SSpin,HIGH);

#else
 
  // Bit bang SPI to ADF7012 
    int i;
    digitalWrite(SCKpin,  LOW );
    delayMicroseconds(2);
    digitalWrite(SSpin, LOW);
    delayMicroseconds(10);

    for(i=31; i>=0; i--) 
    {
        if((data & (unsigned long)(1UL<<i))>>i)
            digitalWrite(MOSIpin, HIGH);
        else
            digitalWrite(MOSIpin, LOW);
        delayMicroseconds(2);
        digitalWrite(SCKpin, HIGH);
        delayMicroseconds(10);
        digitalWrite(SCKpin,  LOW );
        delayMicroseconds(10);
    }
    delayMicroseconds(2);
    digitalWrite(SSpin, HIGH);
#endif
}

//-----------------------------------------------------------
int RadioAdf7012::adf_lock(void) 
{
    // f iddle around with bias and adjust capacity until the vco locks
    
    int adj = adf_config.r0.vco_adjust; // use default start values from setup
    int bias = adf_config.r3.vco_bias;  // or the updated ones that worked last time
    adf_config.r3.pll_enable = 1;
    adf_config.r3.muxout = ADF_MUXOUT_DIGITAL_LOCK;
    adf_write_config();
    delay(50);
    adf_locked();          // passe 1    recup  680
   
    while(!adf_locked()) 
      {
        baltraceln(70,"notlock ",adj); 
        baltraceln(71,"bias ",bias); 
        adf_config.r0.vco_adjust = adj;
        adf_config.r3.vco_bias = bias;
        adf_config.r3.muxout = ADF_MUXOUT_DIGITAL_LOCK;
        adf_write_config();
        delay(50);
        if(++bias == 14) 
        {
            bias = 1;
            if(++adj == 4) 
            {
                baltracesln(72,"Notbias "); 
                adf_config.r0.vco_adjust = 0;
                adf_config.r3.vco_bias = 5;
                
                return NOK;
            }
        }
    }
    baltracesln(73,"Locked"); 
    return OK;
}

//---------------------------------------
int RadioAdf7012::adf_locked(void)
{
 int adc;
  adc = analogRead(ADMUXADFPIN);
  delay(5);
  adc = analogRead(ADMUXADFPIN);
  baltraceln(74,"Lock ",adc); 
  //adc =600; // forçage test
  delay(100);
  if (adc > 500)  { return 1; }  else  { return 0; }
}

//-----------------------------------------------------------
void RadioAdf7012::set_freq(unsigned long freq)
{ 
  // Set the output divider according to recommended ranges given in ADF7012 datasheet  
  // 2012-08-10 TK lowered the borders a bit in order to keep n high enough for 144, 222 and 430 MHz amateur bands 
  // with a constant crystal divider of 8
  adf_config.r0.output_divider = ADF_OUTPUT_DIVIDER_BY_1;
  
  if (freq < 410000000) { adf_config.r0.output_divider = ADF_OUTPUT_DIVIDER_BY_2; };   // cas ici
 // if (freq < 210000000) { adf_config.r0.output_divider = ADF_OUTPUT_DIVIDER_BY_4; }; // cas avec hack externe du ADF
 // if (freq < 130000000) { adf_config.r0.output_divider = ADF_OUTPUT_DIVIDER_BY_8; };

/////   note f5bez  comme la réf  est de 8Mhz les freq out  de valeur entières sont contraintes   on n'aura pas forcément la valeur programmée
////   avec 8Mhz  le step mini est de 244,14hz   8/2p15 32768          
/////  ici l'algo n'est pas optimum sur l'arrondi parte décimale > ou < 0,5 pour  recalcul  N fract  
/////  de même il semble qu'il y a des erreurs par manque au prg origine  
////    THE N-VALUE CHOSEN IS A MINIMUM OF P2 + 3P + 3.         // ????
/////   FOR PRESCALER 8/9 THIS MEANS A MINIMUM N-DIVIDE OF 91.              91  et +   8/9
/////   FOR PRESCALER 4/5 THIS MEANS A MINIMUM N-DIVIDE OF 31.              31  et +   4/5   et en dessous ??
/////    adf_config.r1.prescaler = ADF_PRESCALER_8_9;                // 8/9 requires an integer_n > 91; 4/5 only requires integer_n > 31
//// le pdf dit pour le fract_m   0-4095  et N  0-255
////         >>>>>>>>> avec l'exemplaire M10 du test  sans modifications de la self du pll  <<<<
////         >>>>>>>>> une gamme de 387Mhz à 428Mhz  est testée . La partie 430..440  
////         >>>>>>>>> nécessite un complément à ces régistres ou mettre une self du pll un peu plus petite
////         >>>>>>>>> s'aider de l'exemple 433Mhz du pdf    

  unsigned long f_pfd = ADF7012_CRYSTAL_FREQ / adf_config.r0.r_divider;
  unsigned int n = (unsigned int)(freq / f_pfd);
  float ratio = (float)freq / (float)f_pfd;
  float rest  = ratio - (float)n;
  unsigned long m = (unsigned long)(rest * 4096); 
  adf_config.r1.integer_n = n;
  //// rajouté       /////   code P1  0 pour 4/5 et 1 pour 8/9
  adf_config.r1.prescaler = ADF_PRESCALER_4_5;   // déf  pas clair si n  entre 1 et 31 !!!!        
  if ( n > 31 )     adf_config.r1.prescaler = ADF_PRESCALER_4_5;          
  if ( n > 91 )     adf_config.r1.prescaler = ADF_PRESCALER_8_9;                
  adf_config.r1.fractional_n = m;
  adf_write_register_one();
   // adf_write_config();
  //  revalider ici pour tests  ctrl  pll lock  pour  F  > 410  ou 433 self changée 
  //  ptt_on();          // visu si ppl lock
 // code enlevé pour gain place
}

//-----------------------------------------------------------
void RadioAdf7012::setup()
{
  baltracesln(75,"ADF setup deb"); 
  pinMode(SCKpin,  OUTPUT);
  pinMode(SSpin,   OUTPUT);
  pinMode(MOSIpin, OUTPUT);
  pinMode(ADF7012_TX_DATA_PIN, OUTPUT);

#if !defined(SPI_BITBANG)  
  // Set up SPI
  SPI.setBitOrder(MSBFIRST);
  SPI.setDataMode(SPI_MODE0);
  SPI.setClockDivider(SPI_CLOCK_DIV32);
  // initialize SPI:
  SPI.begin();
#endif 
 
  adf_reset_config(); 
  set_freq(FREQUENCEDEFM10); // Set the default frequency
  adf_write_config();
  digitalWrite(ADF7012_TX_DATA_PIN, LOW);
// personnalisation adjuste F réelle  cf  le 8Mhz 
// 400M demandé  xtal 8Mh du Uc M10   entendu 399980  donc -29kcs
// step 8Mhz / 2 p 15 ou /32768   pas 244,14H  
// adjust  entendu  -20kcs  dessous  donc rajouté  pas = 81,92hz             
   adf_setfecorrection(Mycorrectqrg); //  ou -10 par exemple 
//  ptt_on();  // essai en plus 
//  digitalWrite(TX_PA_PIN, HIGH);  // HIGH = off => make sure the PA is off after boot.
  delay(100);
  
  //Serial.println("ADFdone"); //7012 setup done");
   baltracesln(76,"ADF setup fin"); 
}

//-----------------------------------------------------------
void RadioAdf7012::ptt_on()
{
  int adc;
  digitalWrite(ADF7012_TX_DATA_PIN, LOW);
  adf_config.r3.pa_enable = 0;
  adf_config.r2.power_amplifier_level = 0;
  adf_config.r3.muxout = ADF_MUXOUT_REG_READY;
  adf_write_config();
  delay(100);
  analogRead(ADMUXADFPIN);   // lecture en 2 fois
  adc = analogRead(ADMUXADFPIN);
  baltraceln(77,"rfl ",adc); 
  //adc = 600; //forçage test
  if  (adc < 500)   // Power is bad
   {  baltraceln(78,"toolow",adc); }
  else       // Power is good apparently
  {
    //contrôle recalcule la fréquence générée
    //unsigned long freq_calculated = (ADF7012_CRYSTAL_FREQ / adf_config.r0.r_divider) * ((float)adf_config.r1.integer_n + ((float)adf_config.r1.fractional_n / 4096.0) * ADF7012_CRYSTAL_FREQ );
    //     expl             8000000  /  1             *     n 54  +   ( ( m  125 ) / 4096 )                    ici 433M                    // code enlevé pour gain place et ram 
    if (adf_lock())
    {
      adf_config.r3.pa_enable = 1;
      adf_config.r2.power_amplifier_level = 63; //63 is max power
      baltracesln(79,"Out-on");
      adf_write_config();
      delay(50);
      analogRead(ADMUXADFPIN);   // lecture en 2 fois
      powerlevel = analogRead(ADMUXADFPIN);
      //powerlevel = 200;  // forçage tests
      baltraceln(80,"HFlu ",powerlevel); //
      if (powerlevel > 255) powerlevel = 255; else ptt_off();
     // code enlevé pour gain place
    }
    
  }
}

//-----------------------------------------------------------
void RadioAdf7012::ptt_off()
{
//  divider_test();
  adf_config.r3.pa_enable = 0;
  adf_config.r2.power_amplifier_level = 0; 
  adf_write_config();
  delay(100);
  digitalWrite(ADF7012_TX_DATA_PIN, LOW);
}
//-----------------------------------------------------------
void RadioAdf7012::lock_test(void)
{
  //code enlevé pour gain place
}

//-----------------------------------------------------------
void RadioAdf7012::divider_test(void)
{
 //code enlevé pour gain place
}
//-----------------------------------------------------------
int RadioAdf7012::get_powerlevel()
{
  return powerlevel;
}

//------------fonctions rajoutées f5bez pour ctrl 
void RadioAdf7012::adf_setpwretat(bool etat)
{
  adf_config.r3.pa_enable = etat;
  adf_write_register_three();
}

//-----------------------------------------------------------
void RadioAdf7012::adf_setpwrlevel(int pwrlevel)
{
    adf_config.r2.power_amplifier_level = pwrlevel;
    adf_write_register_two();
}

//-----------------------------------------------------------
void RadioAdf7012::adf_setmodultype(byte modtype)
{

   adf_config.r2.mod_control = modtype;             
   adf_write_register_two();
}

//-----------------------------------------------------------
void RadioAdf7012::adf_setmodullevel(int modlevel)
{
  adf_config.r2.modulation_deviation = (unsigned int)modlevel;    
  adf_write_register_two();
}

//-----------------------------------------------------------
void RadioAdf7012::adf_setfecorrection(int valfcorrect)
{
   adf_config.r0.frequency_error_correction = valfcorrect;   
   adf_write_register_zero();
}

//-----------------------------------------------------------
void RadioAdf7012::adf_setfreqnow(unsigned long freq)
{
   set_freq(freq);
}
//-----------------------------------------------------------

// références :
//https://kt5tk.wordpress.com/2013/12/01/old-version-of-pecan-with-adf7012/
//https://github.com/tkrahn/PecanPico/blob/master/config.h
//https://github.com/tkrahn/PecanPico
// PecanPico APRS Tracker for balloons based on a very old version of Trackuino with a lot of modifications by KT5TK

// Dominique LEVEQUE F5BEZ 
// ==============================================================================================================
